#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* espao para string do tipo "nnn mmm ppp" */
#define STR_ORDEM (3+1+3+1+3+1)

/* Constantes */
#define SIM     1
#define NAO     0

/* Estrutura para Ordem de matrizes
 *      linres x collin = ordem de primeira matriz
 *      collin x colres = ordem de segunda matriz
 */
struct ordem {
  int linres, collin, colres;
};

typedef struct ordem Ordem;

/* Estrutura para grupo de matrizes */
struct matrizes {
  Ordem nordem;
  double **mat_01;
  double **mat_02;
  double **result;
};

typedef struct matrizes Matrizes;

/* ----------- FUNOES ---------------- */

/**********************************************************************
 *  Funcao imprimeMatriz
 * 
 *  Acao:       Imprime o conteudo de uma matriz em stdout
 *  Entrada:    ponteiro para matriz, ordem da matriz (linhas e colunas)
 *  Saida:      nenhuma
 *  Suposicoes: ponteiro para matriz deve ser nao nulo
 **********************************************************************
 */

void imprimeMatriz (double **matriz, int linhas, int colunas)
{
  int i, k;
  for (i=0; i < linhas; i++) {
    for (k=0; k < colunas; k++)
      printf("%5.2lf\t", matriz[i][k]);
    printf("\n");
  }
}

/**********************************************************************
 *  Funcao obtemOrdemMatriz
 * 
 *  Acao:       Obtem de um arquivo a ordem das matrizes que serao 
 *              lidas postriormente pela funcao "obtemMatrizes()"
 *  Entrada:    fp = descritor de arquivo
 *  Saida:      nordem = ordem das matrizes
 *  Retorno:    1 se leitura bem sucedida
 *              0 se chegou ao final do arquivo ou erro na ordem da matriz
 *  Suposicoes: No arquivo de entrada existem 2 ou 3 numeros para a ordem.
 *              O primeiro eh linres, o segundo eh collin. O terceiro eh 
 *              opcional e representa colres.
 *              Na ausencia do terceiro numero, assume-se que colres = linres
 **********************************************************************
 */

int obtemOrdemMatriz (FILE *fp, Ordem *nordem)
{
  char ordem[STR_ORDEM], *str, *token;
  int i,k, ok;

  nordem->linres = nordem->collin = nordem->colres = 0;

  if (ok = fgets(ordem, STR_ORDEM, fp) != NULL)
  {
#ifdef DEBUG
    ordem[strlen(ordem)-1] = '\0';
    printf("MATRIZ %s = ", ordem);
#endif
    
    for (i = 0, token = str = ordem; token; i++, str = NULL){
      token = strtok(str, " ");
      if (token)
	switch (i) {
	case 0: 
	  if (!(nordem->linres = strtol(token, NULL, 10)))
	    token = NULL;
	  break;
	case 1: 
	  if (!(nordem->collin = strtol(token, NULL, 10)))
	    token = NULL;
	  break;
	case 2: 
	  if (!(nordem->colres = strtol(token, NULL, 10)))
	    token = NULL;
	  break;
	}
    }
    
    if (!nordem->linres || !nordem->collin) {
      fprintf(stderr, "Ordem da matriz invlida: %s.\n", ordem);
      ok = NAO;
    }
    else if (!nordem->colres)
      nordem->colres = nordem->linres;

#ifdef DEBUG
    printf("%d x %d x %d\n", nordem->linres, nordem->collin, nordem->colres);
#endif
  }

  return (ok);
}


/**********************************************************************
 *  Funcao obtemMatriz
 * 
 *  Acao:       Obtem do usuario matriz linhasXcolunas de um arquivo
 *  Entrada:    fp = descritor de arquivo
 *              linhas e colunas = ordem da matriz
 *  Retorno:    ponteiro para matriz.
 *              Se NULO, erro em alocacao de memoria
 *  Algoritmo:  espaco para matriz eh alocado com "calloc()"
 **********************************************************************
 */

double **obtemMatriz (FILE *fp, int linhas, int colunas)
{
  int i, k;
  double **matriz;

  if (matriz = (double **) calloc (linhas, sizeof(double *)))
    for (i = 0; matriz && i < linhas; i++) {
      if (matriz[i] = (double *) calloc (colunas, sizeof (double)))
	for (k = 0; k < colunas ; k++)
	  fscanf(fp, "%lf", &matriz[i][k]);
      else
	matriz = NULL;
    }

  return (matriz);
}

/**********************************************************************
 *  Funcao obtemMatrizes
 * 
 *  Acao:       Obtem do usuario matrizes de um arquivo
 *  Entrada:    fp = descritor de arquivo
 *              matrizes = estrutura contendo ordem de matrizes e ponteiros 
 *                         para matrizes que serao lidas do arquivo.
 *  Retorno:    1 se leitura bem sucedida
 *              0 caso contrario
 *  Algoritmo:  espaco para matrizes lidas eh alocado com "calloc()"
 **********************************************************************
 */

int obtemMatrizes (FILE *fp, Matrizes *matrizes)
{
  int i, k, ok = NAO;

  if (matrizes->mat_01 = obtemMatriz (fp, matrizes->nordem.linres,
				      matrizes->nordem.collin))
    if (matrizes->mat_02 = obtemMatriz (fp, matrizes->nordem.collin,
					matrizes->nordem.colres))
      ok = SIM;
  
  return (ok);
}

/**********************************************************************
 *  Funcao multMatrizes
 * 
 *  Acao:       Efetua multiplicacao entre duas matrizes
 *  Entrada:    mult = estrutura com matrizes a serem multiplicadas,
 *                     suas ordens, e ponteiro para resultado
 *  Retorno:    1 se gerou resultado, 
 *              0 caso contrario
 *  Algoritmo:  espaco para matriz com resultado eh alocada com "calloc()"
 **********************************************************************
 */

int multMatrizes (Matrizes *mult)
{
  int i, j, p, ok = NAO;

  /* Cria e inicia em 0 (zero) a matriz que receber o resultado 
   * da multiplicaao entre mult->mat_01 e mult->mat_02
   */
  if (mult->result = (double **) calloc (mult->nordem.linres, sizeof(double *)))
    for (i=0, ok = SIM; ok && i < mult->nordem.linres; i++)
      if (!(mult->result[i] = (double*) calloc (mult->nordem.colres, sizeof(double))))
	ok = NAO;
    
  /* Efetua a multiplicaao */
  if (ok)
    for (i=0; i < mult->nordem.linres; i++)
      for (j=0; j < mult->nordem.colres; j++)
	for (p=0; p < mult->nordem.collin; p++)
	  mult->result[i][j] += mult->mat_01[i][p] * mult->mat_02[p][j];
  else
    fprintf (stderr, "Erro na multiplicao das matrizes\n");

  return (ok);
}

/**********************************************************************
 *  Funcao salvaResult
 * 
 *  Acao:       Salva matriz e sua ordem em arquivo
 *  Entrada:    file = nome do arquivo de entrada
 *              matfinal = estrutura com matriz resultado e sua ordem
 *  Retorno:    1 se salvou com sucesso
 *              0 caso contrario
 **********************************************************************
 */

int salvaResult (char * file, Matrizes *matfinal)
{
  int i, k, ok;
  FILE *fpsai;
  char *arqsai;

  /* Gera nome do arquivo de saida = "res-" + nome arq. entrada */
  arqsai = strdup("res-");
  arqsai = realloc (arqsai, strlen(arqsai) + strlen(file) + 1);
  arqsai = strcat (arqsai, file);

  /* Abre arquivo de saida para anexar e grava resultado */
  if (ok = (fpsai = fopen (arqsai, "a")) != NULL) {
    fprintf(fpsai, "%d %d\n", matfinal->nordem.linres, matfinal->nordem.colres);
    for (i=0; i < matfinal->nordem.linres; i++) {
      for (k=0; k < matfinal->nordem.colres; k++)
	fprintf(fpsai, "%5.2lf ", matfinal->result[i][k]);
      fprintf(fpsai, "\n");
    }
    
    fclose (fpsai);
  }
  else
    fprintf (stderr, "Erro na abertura do arq. \"%s\"\n", arqsai);

  return (ok);
}

/**********************************************
 * Programa principal
 * Forma de uso: prog [arquivo]
 *               se "arquivo" nao eh fornecido,
 *               assume "matriz_01.dat"
 **********************************************
 */

main (int argc, char *argv[]) 
{
  char *arqent = "matriz_02.dat";
  FILE *fpent;

  Matrizes mats = {{0,0,0}, NULL, NULL, NULL};;

  if (argc == 2)
    arqent = argv[1];
  
  if (!(fpent = fopen (arqent, "r")))
    fprintf (stderr, "Erro na abertura do arq. \"%s\"\n", arqent);
  else {
    while (obtemOrdemMatriz(fpent, &mats.nordem))
    {
      if (! obtemMatrizes (fpent, &mats))
	exit (1);
      else if (! multMatrizes(&mats))
	exit (2);
      else if (! salvaResult(arqent, &mats))
	exit (3);

      /* Flush fpent*/
      while (fgetc(fpent) != '\n');
    }
    fclose (fpent);
  }
}

